"""
Builder to generate elasticity docs.

The build proceeds in the below steps:
1. Use materials builder to group tasks according the formula, space group,
   structure matching
2. Filter opt and deform tasks using calc type
3. Filter opt and deform tasks to match prescribed INCAR params
4. Group opt and deform tasks by parent lattice, i.e. lattice before deformation
5. For each group, select the one with the latest completed time (all tasks in a
   group are regarded as the same after going through all the filters)
6. For all opt-deform tasks groups with the same parent lattice, select the group with
   the most number of deformation tasks as the final data fot fitting the elastic tensor
7. Fit the elastic tensor
"""

from datetime import datetime
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

import numpy as np
from maggma.core import Builder, Store
from pydash.objects import get
from pymatgen.analysis.elasticity.strain import Deformation
from pymatgen.analysis.elasticity.stress import Stress
from pymatgen.core import Structure
from pymatgen.core.tensors import TensorMapping

from emmet.core.elasticity import ElasticityDoc
from emmet.core.mpid import MPID
from emmet.core.utils import jsanitize
from emmet.core.vasp.calc_types import CalcType


class ElasticityBuilder(Builder):
    def __init__(
        self,
        tasks: Store,
        materials: Store,
        elasticity: Store,
        query: Optional[Dict] = None,
        fitting_method: str = "finite_difference",
        **kwargs,
    ):
        """
        Creates an elastic collection for materials.

        Args:
            tasks: Store of tasks
            materials: Store of materials
            elasticity: Store of elasticity
            query: Mongo-like query to limit the tasks to be analyzed
            fitting_method: method to fit the elastic tensor: {`finite_difference`,
                `pseudoinverse`, `independent`}
        """

        self.tasks = tasks
        self.materials = materials
        self.elasticity = elasticity
        self.query = query if query is not None else {}
        self.fitting_method = fitting_method
        self.kwargs = kwargs

        super().__init__(sources=[tasks, materials], targets=[elasticity], **kwargs)

    def ensure_index(self):
        self.tasks.ensure_index("nsites")
        self.tasks.ensure_index("formula_pretty")
        self.tasks.ensure_index("last_updated")

        self.materials.ensure_index("material_id")
        self.materials.ensure_index("last_updated")

        self.elasticity.ensure_index("material_id")
        self.elasticity.ensure_index("last_updated")

    def get_items(
        self,
    ) -> Generator[Tuple[str, Dict[str, str], List[Dict]], None, None]:
        """
        Gets all items to process into elasticity docs.

        Returns:
            material_id: material id for the tasks
            calc_types: calculation types of the tasks
            tasks: task docs belong to the same material
        """

        self.logger.info("Elastic Builder Started")

        self.ensure_index()

        cursor = self.materials.query(
            criteria=self.query, properties=["material_id", "calc_types", "task_ids"]
        )

        # query for tasks
        query = self.query.copy()

        for i, doc in enumerate(cursor):

            material_id = doc["material_id"]
            calc_types = {str(k): v for k, v in doc["calc_types"].items()}

            self.logger.debug(f"Querying tasks for material {material_id} (index {i}).")

            # update query with task_ids
            query["task_id"] = {"$in": [int(i) for i in doc["task_ids"]]}

            projections = [
                "output",
                "orig_inputs",
                "completed_at",
                "transmuter",
                "task_id",
                "dir_name",
            ]

            task_cursor = self.tasks.query(criteria=query, properties=projections)
            tasks = list(task_cursor)

            yield material_id, calc_types, tasks

    def process_item(
        self, item: Tuple[MPID, Dict[str, str], List[Dict]]
    ) -> Union[Dict, None]:
        """
        Process all tasks belong to the same material into an elasticity doc.

        Args:
            item:
                material_id: material id for the tasks
                calc_types: {task_id: task_type} calculation types of the tasks
                tasks: task docs belong to the same material

        Returns:
            Elasticity doc obtained from the list of tasks. `None` if failed to
            obtain the elasticity doc from the tasks.
        """

        material_id, calc_types, tasks = item

        if len(tasks) != len(calc_types):
            self.logger.error(
                f"Number of tasks ({len(tasks)}) is not equal to number of calculation "
                f"types ({len(calc_types)}) for material with material id "
                f"{material_id}. Cannot proceed."
            )
            return None

        # filter by calc type
        opt_tasks = filter_opt_tasks(tasks, calc_types)
        deform_tasks = filter_deform_tasks(tasks, calc_types)
        if not opt_tasks or not deform_tasks:
            return None

        # filter by incar
        opt_tasks = filter_by_incar_settings(opt_tasks)
        deform_tasks = filter_by_incar_settings(deform_tasks)
        if not opt_tasks or not deform_tasks:
            return None

        # select one task for each set of optimization tasks with the same lattice
        opt_grouped_tmp = group_by_parent_lattice(opt_tasks, mode="opt")
        opt_grouped = [
            (lattice, filter_opt_tasks_by_time(tasks, self.logger))
            for lattice, tasks in opt_grouped_tmp
        ]

        # for deformed tasks with the same lattice, select one if there are multiple
        # tasks with the same deformation
        deform_grouped = group_by_parent_lattice(deform_tasks, mode="deform")
        deform_grouped = [
            (lattice, filter_deform_tasks_by_time(tasks))
            for lattice, tasks in deform_grouped
        ]

        # select opt and deform tasks for fitting
        final_opt, final_deform = select_final_opt_deform_tasks(
            opt_grouped, deform_grouped, self.logger
        )
        if final_opt is None or final_deform is None:
            return None

        # convert to elasticity doc
        deforms = []
        stresses = []
        deform_task_ids = []
        deform_dir_names = []
        for doc in final_deform:
            deforms.append(
                Deformation(
                    doc["transmuter"]["transformation_params"][0]["deformation"]
                )
            )
            # -0.1 to convert to GPa from kBar and s
            stresses.append(-0.1 * Stress(doc["output"]["stress"]))
            deform_task_ids.append(doc["task_id"])
            deform_dir_names.append(doc["dir_name"])

        elasticity_doc = ElasticityDoc.from_deformations_and_stresses(
            structure=Structure.from_dict(final_opt["output"]["structure"]),
            material_id=material_id,
            deformations=deforms,
            stresses=stresses,
            deformation_task_ids=deform_task_ids,
            deformation_dir_names=deform_dir_names,
            equilibrium_stress=-0.1 * Stress(final_opt["output"]["stress"]),
            optimization_task_id=final_opt["task_id"],
            optimization_dir_name=final_opt["dir_name"],
            fitting_method="finite_difference",
        )
        elasticity_doc = jsanitize(elasticity_doc.dict(), allow_bson=True)

        return elasticity_doc

    def update_targets(self, items: List[Dict]):
        """
        Insert the new elasticity docs into the elasticity collection.

        Args:
            items: elasticity docs
        """
        self.logger.info(f"Updating {len(items)} elasticity documents")

        self.elasticity.update(items, key="material_id")


def filter_opt_tasks(
    tasks: List[Dict],
    calc_types: Dict[str, str],
    target_calc_type: str = CalcType.GGA_Structure_Optimization,
) -> List[Dict]:
    """
    Filter optimization tasks, by
        - calculation type
    """
    opt_tasks = [t for t in tasks if calc_types[str(t["task_id"])] == target_calc_type]

    return opt_tasks


def filter_deform_tasks(
    tasks: List[Dict],
    calc_types: Dict[str, str],
    target_calc_type: str = CalcType.GGA_Deformation,
) -> List[Dict]:
    """
    Filter deformation tasks, by
        - calculation type
        - number of transformations
        - transformation class
    """
    deform_tasks = []
    for t in tasks:
        if calc_types[str(t["task_id"])] == target_calc_type:
            transforms = t["transmuter"]["transformations"]
            if (
                len(transforms) == 1
                and transforms[0] == "DeformStructureTransformation"
            ):
                deform_tasks.append(t)

    return deform_tasks


def filter_by_incar_settings(
    tasks: List[Dict], incar_settings: Dict[str, Any] = None
) -> List[Dict]:
    """
    Filter tasks by incar parameters.
    """

    if incar_settings is None:
        incar_settings = {
            "LREAL": False,
            "ENCUT": 700,
            "PREC": "Accurate",
            "EDIFF": 1e-6,
        }

    selected = []
    for t in tasks:
        incar = t["orig_inputs"]["incar"]
        ok = True
        for k, v in incar_settings.items():
            if k not in incar:
                ok = False
                break

            if isinstance(incar[k], str):
                if incar[k].lower() != v.lower():
                    ok = False
                    break

            elif isinstance(incar[k], float):
                if not np.allclose(incar[k], v, atol=1e-10):
                    ok = False
                    break

            else:
                if incar[k] != v:
                    ok = False
                    break

        if ok:
            selected.append(t)

    return selected


def filter_opt_tasks_by_time(tasks: List[Dict], logger) -> Dict:
    """
    Filter a set of tasks to select the latest completed one.

    Args:
        tasks: the set of tasks to filer
        logger:

    Returns:
        selected latest task
    """
    if len(tasks) == 0:
        raise RuntimeError("Cannot select latest from 0 tasks")
    elif len(tasks) == 1:
        return tasks[0]
    else:
        completed = [(datetime.fromisoformat(t["completed_at"]), t) for t in tasks]
        sorted_by_completed = sorted(completed, key=lambda pair: pair[0])
        latest_pair = sorted_by_completed[-1]
        selected = latest_pair[1]

        task_ids = [t["task_id"] for t in tasks]
        logger.warning(
            f"Select the latest optimization task {selected['task_id']} completed at "
            f"{selected['completed_at']} from a set of tasks: {task_ids}."
        )

        return selected


def filter_deform_tasks_by_time(
    tasks: List[Dict], deform_comp_tol: float = 1e-5
) -> List[Dict]:
    """
    For deformation tasks with the same deformation, select the latest completed one.

    Args:
        tasks: the deformation tasks
        deform_comp_tol: tolerance for comparing deformation equivalence

    Returns:
        filtered deformation tasks
    """

    mapping = TensorMapping(tol=deform_comp_tol)

    for doc in tasks:

        # assume only one deformation, should be checked in `filter_deform_tasks()`
        deform = doc["transmuter"]["transformation_params"][0]["deformation"]

        if deform in mapping:
            current = datetime.fromisoformat(doc["completed_at"])
            exist = datetime.fromisoformat(mapping[deform]["completed_at"])
            if current > exist:
                mapping[deform] = doc
        else:
            mapping[deform] = doc

    selected = list(mapping.values())

    return selected


def select_final_opt_deform_tasks(
    opt_tasks: List[Tuple[np.ndarray, Dict]],
    deform_tasks: List[Tuple[np.ndarray, List[Dict]]],
    logger,
    lattice_comp_tol: float = 1e-5,
) -> Tuple[Union[Dict, None], Union[List[Dict], None]]:
    """
    Select the final opt task and deform tasks for fitting.

    This is achieved by selecting the opt--deform pairs with the same lattice,
    and also with the most deform tasks.

    Returns:
        final_opt_task: selected opt task
        final_deform_tasks: selected deform tasks
    """

    # group opt and deform tasks by lattice
    mapping = TensorMapping(tol=lattice_comp_tol)
    for lat, ot in opt_tasks:
        mapping[lat] = {"opt_task": ot}

    for lat, dt in deform_tasks:
        if lat in mapping:
            mapping[lat]["deform_tasks"] = dt
        else:
            mapping[lat] = {"deform_tasks": dt}

    # select opt--deform paris with the most deform tasks
    selected = None
    num_deform_tasks = -1
    for lat, tasks in mapping.items():
        if "opt_task" in tasks and "deform_tasks" in tasks:
            n = len(tasks["deform_tasks"])
            if n > num_deform_tasks:
                selected = (tasks["opt_task"], tasks["deform_tasks"])
                num_deform_tasks = n

    if selected is None:
        tasks = [pair[1] for pair in opt_tasks]
        for pair in deform_tasks:
            tasks.extend(pair[1])

        ids = [t["task_id"] for t in tasks]
        logger.warning(
            f"Cannot find optimization and deformation tasks that match by lattice "
            f"for tasks {ids}"
        )

        final_opt_task = None
        final_deform_tasks = None
    else:
        final_opt_task, final_deform_tasks = selected

    return final_opt_task, final_deform_tasks


def group_by_parent_lattice(
    tasks: List[Dict], mode: str, lattice_comp_tol: float = 1e-5
) -> List[Tuple[np.ndarray, List[Dict]]]:
    """
    Groups a set of task docs by parent lattice equivalence.

    Args:
        tasks: task docs
        mode: determines which lattice to use. If `opt`, use the lattice of the
            output structure, and this is intended for optimization tasks. If
            `deform`, use the lattice of the output structure and transform it by the
            deformation in transmuter, and this is intended for deformation tasks.
        lattice_comp_tol: tolerance for comparing lattice equivalence.

    Returns:
        [(lattice, List[tasks])]: each tuple gives the common parent lattice of a
            list of the structures before deformation (if any), and the list tasks
            from which the structures are taken.
    """
    docs_by_lattice: List[Tuple[np.ndarray, List[Dict]]] = []

    for doc in tasks:
        sim_lattice = get(doc, "output.structure.lattice.matrix")

        if mode == "deform":
            transform_params = doc["transmuter"]["transformation_params"]
            deform = transform_params[0]["deformation"]
            parent_lattice = np.dot(sim_lattice, np.transpose(np.linalg.inv(deform)))
        elif mode == "opt":
            parent_lattice = np.array(sim_lattice)
        else:
            raise ValueError(f"Unsupported mode {mode}")

        match = False
        for unique_lattice, lattice_docs in docs_by_lattice:
            match = np.allclose(unique_lattice, parent_lattice, atol=lattice_comp_tol)
            if match:
                lattice_docs.append(doc)
                break
        if not match:
            docs_by_lattice.append((parent_lattice, [doc]))

    return docs_by_lattice



from datetime import datetime
from itertools import chain
from math import ceil
from typing import Optional, Iterable, Iterator, List, Dict

from maggma.builders import Builder
from maggma.core import Store
from maggma.utils import grouper

from emmet.core.qchem.task import TaskDocument
from emmet.core.qchem.molecule import MoleculeDoc, evaluate_lot
from emmet.core.molecules.atomic import (
    PartialChargesDoc,
    PartialSpinsDoc,
    CHARGES_METHODS,
    SPINS_METHODS,
)
from emmet.core.utils import jsanitize
from emmet.builders.settings import EmmetBuildSettings


__author__ = "Evan Spotte-Smith"

SETTINGS = EmmetBuildSettings()


class PartialChargesBuilder(Builder):
    """
    The PartialChargesBuilder extracts partial charges data from a MoleculeDoc.

    Various methods can be used to define partial charges, including:
        - Mulliken
        - Restrained Electrostatic Potential (RESP)
        - Critic2
        - Natural Bonding Orbital (NBO) population analysis

    This builder will attempt to build documents for each molecule with each method.
    For each molecule-method combination, the highest-quality data available (based
    on level of theory and electronic energy) will be used.

    The process is as follows:
        1. Gather MoleculeDocs by formula
        2. For each molecule, sort all associated tasks by level of theory and electronic energy
        2. For each method:
            2.1. Find task docs with necessary data to calculate partial charges by that method
            2.2. Take best (defined by level of theory and electronic energy) task
            2.3. Convert TaskDoc to PartialChargesDoc
    """

    def __init__(
        self,
        tasks: Store,
        molecules: Store,
        charges: Store,
        query: Optional[Dict] = None,
        methods: Optional[List] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):

        self.tasks = tasks
        self.molecules = molecules
        self.charges = charges
        self.query = query if query else dict()
        self.methods = methods if methods else CHARGES_METHODS
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        super().__init__(sources=[tasks, molecules], targets=[charges])

    def ensure_indexes(self):
        """
        Ensures indices on the collections needed for building
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id")
        self.tasks.ensure_index("last_updated")
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_alphabetical")

        # Search index for molecules
        self.molecules.ensure_index("molecule_id")
        self.molecules.ensure_index("last_updated")
        self.molecules.ensure_index("task_ids")
        self.molecules.ensure_index("formula_alphabetical")

        # Search index for charges
        self.charges.ensure_index("molecule_id")
        self.charges.ensure_index("method")
        self.charges.ensure_index("task_id")
        self.charges.ensure_index("last_updated")
        self.charges.ensure_index("formula_alphabetical")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the builder for distributed computation"""

        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.charges.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        N = ceil(len(to_process_forms) / number_splits)

        for formula_chunk in grouper(to_process_forms, N):
            yield {"query": {"formula_alphabetical": {"$in": list(formula_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into partial charges documents.
        This does no datetime checking; relying on on whether
        task_ids are included in the charges Store

        Returns:
            generator or list relevant tasks and molecules to process into documents
        """

        self.logger.info("Partial charges builder started")
        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime
        self.timestamp = datetime.utcnow()

        # Get all processed molecules
        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.charges.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        self.logger.info(f"Found {len(to_process_docs)} unprocessed documents")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        for formula in to_process_forms:
            mol_query = dict(temp_query)
            mol_query["formula_alphabetical"] = formula
            molecules = list(self.molecules.query(criteria=mol_query))

            yield molecules

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process the tasks into PartialChargesDocs

        Args:
            tasks List[Dict] : a list of MoleculeDocs in dict form

        Returns:
            [dict] : a list of new partial charges docs
        """

        mols = [MoleculeDoc(**item) for item in items]
        formula = mols[0].formula_alphabetical
        mol_ids = [m.molecule_id for m in mols]
        self.logger.debug(f"Processing {formula} : {mol_ids}")

        charges_docs = list()

        for mol in mols:
            correct_charge_spin = [
                e
                for e in mol.entries
                if e["charge"] == mol.charge
                and e["spin_multiplicity"] == mol.spin_multiplicity
            ]

            sorted_entries = sorted(
                correct_charge_spin,
                key=lambda x: (sum(evaluate_lot(x["level_of_theory"])), x["energy"]),
            )

            for method in self.methods:
                # For each method, grab entries that have the relevant data
                relevant_entries = [
                    e
                    for e in sorted_entries
                    if e.get(method) is not None or e["output"].get(method) is not None
                ]

                if len(relevant_entries) == 0:
                    continue

                # Grab task document of best entry
                best_entry = relevant_entries[0]
                task = best_entry["task_id"]

                task_doc = TaskDocument(**self.tasks.query_one({"task_id": int(task)}))

                doc = PartialChargesDoc.from_task(
                    task_doc,
                    molecule_id=mol.molecule_id,
                    preferred_methods=[method],
                    deprecated=False,
                )

                charges_docs.append(doc)

        self.logger.debug(f"Produced {len(charges_docs)} charges docs for {formula}")

        return jsanitize([doc.dict() for doc in charges_docs], allow_bson=True)

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new documents into the charges collection

        Args:
            items [[dict]]: A list of documents to update
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        # Add timestamp
        for item in docs:
            item.update(
                {
                    "_bt": self.timestamp,
                }
            )

        molecule_ids = list({item["molecule_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} partial charges documents")
            self.charges.remove_docs({self.charges.key: {"$in": molecule_ids}})
            # Neither molecule_id nor method need to be unique, but the combination must be
            self.charges.update(
                docs=docs,
                key=["molecule_id", "method"],
            )
        else:
            self.logger.info("No items to update")


class PartialSpinsBuilder(Builder):
    """
    The PartialSpinsBuilder extracts partial spin data from a MoleculeDoc.

    Various methods can be used to define partial atomic spins, including:
        - Mulliken
        - Natural Bonding Orbital (NBO) population analysis

    This builder will attempt to build documents for each molecule with each method.
    For each molecule-method combination, the highest-quality data available (based
    on level of theory and electronic energy) will be used.

    The process is as follows:
        1. Gather MoleculeDocs by formula
        2. For each molecule, sort all associated tasks by level of theory and electronic energy
        2. For each method:
            2.1. Find task docs with necessary data to calculate partial spins by that method
            2.2. Take best (defined by level of theory and electronic energy) task
            2.3. Convert TaskDoc to PartialChargesDoc
    """

    def __init__(
        self,
        tasks: Store,
        molecules: Store,
        spins: Store,
        query: Optional[Dict] = None,
        methods: Optional[List] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):

        self.tasks = tasks
        self.molecules = molecules
        self.spins = spins
        self.query = query if query else dict()
        self.methods = methods if methods else SPINS_METHODS
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        super().__init__(sources=[tasks, molecules], targets=[spins])

    def ensure_indexes(self):
        """
        Ensures indices on the collections needed for building
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id")
        self.tasks.ensure_index("last_updated")
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_alphabetical")

        # Search index for molecules
        self.molecules.ensure_index("molecule_id")
        self.molecules.ensure_index("last_updated")
        self.molecules.ensure_index("task_ids")
        self.molecules.ensure_index("formula_alphabetical")

        # Search index for charges
        self.spins.ensure_index("molecule_id")
        self.spins.ensure_index("method")
        self.spins.ensure_index("task_id")
        self.spins.ensure_index("last_updated")
        self.spins.ensure_index("formula_alphabetical")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the builder for distributed computation"""

        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.spins.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        N = ceil(len(to_process_forms) / number_splits)

        for formula_chunk in grouper(to_process_forms, N):
            yield {"query": {"formula_alphabetical": {"$in": list(formula_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into partial spins documents.
        This does no datetime checking; relying on on whether
        task_ids are included in the spins Store

        Returns:
            generator or list relevant tasks and molecules to process into documents
        """

        self.logger.info("Partial spins builder started")
        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime
        self.timestamp = datetime.utcnow()

        # Get all processed molecules
        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.spins.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        self.logger.info(f"Found {len(to_process_docs)} unprocessed documents")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        for formula in to_process_forms:
            mol_query = dict(temp_query)
            mol_query["formula_alphabetical"] = formula
            molecules = list(self.molecules.query(criteria=mol_query))

            yield molecules

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process the tasks into PartialSpinsDocs

        Args:
            tasks List[Dict] : a list of MoleculeDocs in dict form

        Returns:
            [dict] : a list of new partial spins docs
        """

        mols = [MoleculeDoc(**item) for item in items]
        formula = mols[0].formula_alphabetical
        mol_ids = [m.molecule_id for m in mols]
        self.logger.debug(f"Processing {formula} : {mol_ids}")

        spins_docs = list()

        for mol in mols:
            # Molecule with spin multiplicity 1 has no partial spins
            if mol.spin_multiplicity == 1:
                continue

            correct_charge_spin = [
                e
                for e in mol.entries
                if e["charge"] == mol.charge
                and e["spin_multiplicity"] == mol.spin_multiplicity
            ]

            sorted_entries = sorted(
                correct_charge_spin,
                key=lambda x: (sum(evaluate_lot(x["level_of_theory"])), x["energy"]),
            )

            for method in self.methods:
                # For each method, grab entries that have the relevant data
                relevant_entries = [
                    e
                    for e in sorted_entries
                    if e.get(method) is not None or e["output"].get(method) is not None
                ]

                if len(relevant_entries) == 0:
                    continue

                # Grab task document of best entry
                best_entry = relevant_entries[0]
                task = best_entry["task_id"]

                task_doc = TaskDocument(**self.tasks.query_one({"task_id": int(task)}))

                doc = PartialSpinsDoc.from_task(
                    task_doc,
                    molecule_id=mol.molecule_id,
                    preferred_methods=[method],
                    deprecated=False,
                )

                spins_docs.append(doc)

        self.logger.debug(
            f"Produced {len(spins_docs)} partial spins docs for {formula}"
        )

        return jsanitize([doc.dict() for doc in spins_docs], allow_bson=True)

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new documents into the spins collection

        Args:
            items [[dict]]: A list of documents to update
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        # Add timestamp
        for item in docs:
            item.update(
                {
                    "_bt": self.timestamp,
                }
            )

        molecule_ids = list({item["molecule_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} partial spins documents")
            self.spins.remove_docs({self.spins.key: {"$in": molecule_ids}})
            # Neither molecule_id nor method need to be unique, but the combination must be
            self.spins.update(
                docs=docs,
                key=["molecule_id", "method"],
            )
        else:
            self.logger.info("No items to update")



from typing import Dict, Optional, List
from maggma.builders.map_builder import MapBuilder
from maggma.core import Store

from pymatgen.core.structure import Structure
from emmet.core.robocrys import RobocrystallogapherDoc
from emmet.core.utils import jsanitize


class RobocrystallographerBuilder(MapBuilder):
    def __init__(
        self,
        oxidation_states: Store,
        robocrys: Store,
        query: Optional[Dict] = None,
        **kwargs
    ):
        self.oxidation_states = oxidation_states
        self.robocrys = robocrys
        self.kwargs = kwargs

        self.robocrys.key = "material_id"
        self.oxidation_states.key = "material_id"

        super().__init__(
            source=oxidation_states,
            target=robocrys,
            query=query,
            projection=["material_id", "structure", "deprecated"],
            **kwargs
        )

    def unary_function(self, item):
        structure = Structure.from_dict(item["structure"])
        mpid = item["material_id"]
        deprecated = item["deprecated"]

        doc = RobocrystallogapherDoc.from_structure(
            structure=structure,
            material_id=mpid,
            deprecated=deprecated,
            fields=[],
        )

        return jsanitize(doc.dict(), allow_bson=True)



from maggma.builders.map_builder import MapBuilder
from maggma.core import Store
from pymatgen.core import Structure
from pymatgen.core import __version__ as pymatgen_version

from emmet.core.oxidation_states import OxidationStateDoc
from emmet.core.utils import jsanitize


class OxidationStatesBuilder(MapBuilder):
    def __init__(
        self,
        materials: Store,
        oxidation_states: Store,
        query=None,
        **kwargs,
    ):
        """
        Creates Oxidation State documents from materials

        Args:
            materials: Store of materials docs
            oxidation_states: Store to update with oxidation state document
            query : query on materials to limit search
        """
        self.materials = materials
        self.oxidation_states = oxidation_states
        self.kwargs = kwargs
        self.query = query or {}

        # Enforce that we key on material_id
        self.materials.key = "material_id"
        self.oxidation_states.key = "material_id"
        super().__init__(
            source=materials,
            target=oxidation_states,
            projection=["structure", "deprecated"],
            query=query,
            **kwargs,
        )

    def unary_function(self, item):
        structure = Structure.from_dict(item["structure"])
        mpid = item["material_id"]
        deprecated = item["deprecated"]

        oxi_doc = OxidationStateDoc.from_structure(
            structure=structure, material_id=mpid, deprecated=deprecated
        )
        doc = jsanitize(oxi_doc.dict(), allow_bson=True)

        return doc
